﻿#Region "FileHeader"
#If False Then
StreamContainer can be set on an endless stream to write several Contents into the BaseStream. 
Mark up Content-End by calling .NextContent().
In Reading-mode StreamContainer only reads one Content, that means, it behaves like an endable Stream. This solves some Problems of reading several Contents from endless Streams.
After reading one Content open the next for reading by calling .NextContent().

Internally StreamContainer inserts "package-Headers" into the BaseStream, which say the length of the following package (and where the next Header is). The value of the last package-Header is negative, since it is bitwise inverted.
So the reading-Process reads the headers and delivers the content-data to output. When a negative header is reached, it will be re-inverted and says the length of the **last** package of this content, and after that no more data will be delivered (until calling .NextContent()).

the writing is optimized to avoid using the _Buffer: Small amounts of data will be buffered, but when a new Write-Demand exceeds the bufsize, then the buffer flushes  into the BaseStream, and moreover the exceeding data will be written **also** into BaseStream - skipping the _Buffer.
In effect Write-Demands with count > 1020 will allways skip the buffering.

in other words: Writing with many small Write-Demands results in packages with minimum-size 1024 (including the header). Bigger Write-Demands generate packages in size of the data (and aditional header-size).
#End If '-- Options, Imports
Imports System.IO
#End Region 'FileHeader

Public Class StreamContainer : Inherits Stream

   Private Const _ReservedOffset As Integer = 4 ' the first _Buffer-bytes are reserved to note down the package-size
   Private Const _WriterBufSize As Integer = 1024
   Private ReadOnly _Buffer As Byte()
   Private _Offset As Integer 'current _Buffer-Write-Position
   Private _HeaderDistance As Integer ' distance to the next Header in the BaseStream
   Private _IsLastPackage As Boolean ' when true no more packages belong to current content. Complete reading the current package, then simulate EOF.

   Private _BaseStream As Stream

   Public Sub New(strm As Stream, access As FileAccess)
      ' Reading-mode needs only a very small Buffer
      _Buffer = New Byte(If(access = FileAccess.Write, _WriterBufSize, _ReservedOffset) - 1) {}
      _BaseStream = strm
      Reset()
   End Sub

   'Note: all Public Overrides can be called from everywhere, including Owning Streams, Readers or Writers of each kind. Especially Read(), Write() and Flush() must behave according to standard-Stream-Behavior.
   Public Overrides Function Read(buffer() As Byte, offset As Integer, count As Integer) As Integer
      If _Offset = 0 Then _Offset = _ReservedOffset 'indicate first Demand after Reset
      Dim offs0 = offset
      If _HeaderDistance = 0 Then ReadReadBytesLeft()
      Dim readCount = Math.Min(_HeaderDistance, count)
      While readCount > 0
         readCount = _BaseStream.Read(buffer, offset, readCount)
         offset += readCount
         count -= readCount
         _HeaderDistance -= readCount
         If _HeaderDistance = 0 Then ReadReadBytesLeft()
         readCount = Math.Min(_HeaderDistance, count)
      End While
      Return offset - offs0
   End Function
   Private Sub ReadReadBytesLeft()
      If _IsLastPackage Then Return
      If _BaseStream.Read(_Buffer, 0, _ReservedOffset) < _ReservedOffset Then Throw New IOException("couldn't read from BaseStream")
      _HeaderDistance = BitConverter.ToInt32(_Buffer, 0)
      If _HeaderDistance < 0 Then _HeaderDistance = Not _HeaderDistance : _IsLastPackage = True
   End Sub

   Public Overrides Sub Write(buffer() As Byte, offset As Integer, count As Integer)
      If _Offset = 0 Then _Offset = _ReservedOffset 'indicate first Demand after Reset
      If count < _Buffer.Length - _Offset Then 'write into _Buffer
         Array.Copy(buffer, offset, _Buffer, _Offset, count)
         _Offset += count
      Else ' flush _Buffer and additional write count Bytes to Stream
         FlushBuf(count)
         _BaseStream.Write(buffer, offset, count)
      End If
   End Sub
   Private Sub FlushBuf(Optional additionalCount As Integer = -1)
      Dim packageSize = _Offset - _ReservedOffset
      If additionalCount < 0 Then : packageSize = Not packageSize ' negative packageSizes indicate last package of Content
      Else : packageSize += additionalCount ' additionalCount Bytes are to write after the FlushBuf-Call
      End If
      Array.Copy(BitConverter.GetBytes(packageSize), _Buffer, _ReservedOffset) ' note down packageSize in reserved _Buffer-Range
      _BaseStream.Write(_Buffer, 0, _Offset) ' then write complete _Buffer to Stream
      _Offset = _ReservedOffset
   End Sub

   Public Sub NextContent()
      If _Offset = 0 Then Return
      If _Buffer.Length = _ReservedOffset Then
         If Not _IsLastPackage Then Throw New InvalidOperationException("cannot initialize next Content when current Content is not read to end")
      Else
         FlushBuf()
         _BaseStream.Flush()
      End If
      Reset()
   End Sub

   Private Sub Reset()
      _IsLastPackage = False
      _Offset = 0
      _HeaderDistance = 0
   End Sub

   <Obsolete("Flush has no effect. Call NextContent to close the current and start an new one")>
   Public Overrides Sub Flush()
   End Sub

   Public Overrides ReadOnly Property CanRead As Boolean
      Get
         CanRead = _BaseStream.CanRead AndAlso _Buffer.Length = _ReservedOffset
      End Get
   End Property

   Public Overrides ReadOnly Property CanSeek As Boolean
      Get
         Return False
      End Get
   End Property

   Public Overrides ReadOnly Property CanWrite As Boolean
      Get
         CanWrite = _BaseStream.CanWrite AndAlso _Buffer.Length > _ReservedOffset
      End Get
   End Property

   Public Overrides ReadOnly Property Length As Long
      Get
         Return _BaseStream.Length
      End Get
   End Property

   Public Overrides Property Position As Long
      Get
         Return _BaseStream.Position
      End Get
      Set(value As Long)
         Throw New NotSupportedException
      End Set
   End Property

   Public Overrides Function Seek(offset As Long, origin As SeekOrigin) As Long
      Throw New NotSupportedException
   End Function

   Public Overrides Sub SetLength(value As Long)
      Throw New NotSupportedException
   End Sub

End Class
